# -*- coding: utf-8 -*-
"""
Created on Thu Dec 15 21:24:09 2022

@author: yrc2
"""
from biorefineries.tea.conventional_ethanol_tea import *

__all__ = ('create_tea', 'ConventionalEthanolTEA')

create_tea = create_conventional_ethanol_tea
