# -*- coding: utf-8 -*-
"""
Created on Tue Feb  4 06:42:02 2020

@author: yoelr
"""
from biorefineries.cane import create_cellulosic_oilcane_chemicals as create_chemicals

__all__ = ('create_chemicals',)


    
