# -*- coding: utf-8 -*-
"""
Created on Tue Feb  4 06:42:02 2020

@author: yoelr
"""
from biorefineries.cane.chemicals import *

__all__ = ('create_chemicals',)

create_chemicals = create_oilcane_chemicals


